/*!
* \file CalculPropagation.cpp
* \brief General Calculations for noise propagation with the NMPB 2008 method
* \author CSTB
* \version 1.0
*/

#include "CalculPropagation.h"
#include "SousCalculs/Diffraction.h"
#include <math.h>
#include <stdio.h>
#include <assert.h>

using namespace std;
using namespace DiffractionNMPB;

namespace CalculPropagationNMPB
{
	
	// ------------------------------ utility function ------------------------------------
	/*!
	 * \brief Calculates function h(x) as defined by formula (B.2) of annexe B
	 *
	 */
	static int modified_floor (double index)
	{
		double x = floor (index) ;
		int i  = (int) x ;
		if (x != index) i++ ;
		return i ;
	}

	// ------------------------------ Sound levels -----------------------------------------
	/*!
	 * \brief Calculates sound levels sum
	 *
	 * Used for  5.2.4 Formula (8) and  5.2.5 Formula (9)
	 * 
	 * \param n the sound levels number
	 * 
	 * \param levels The sound levels to sum
	 * 
	 * \returns the sum
	 */
	double SumLevels(int n, double const* levels)
	{
		double soundsSum = 0;
		for(int i = 0 ; i < n ; i++)
		{
			soundsSum += pow(10, 0.1 * levels[i]);
		}

		double soundsSummed = 10 * log10(soundsSum);

		return soundsSummed;
	}

	/*!
	 * \brief Calculates long-term sound level for the given the sound levels (homogeneous and favorable) for a path
	 *
	 * p.18 -  5.2.3 Formula (7)
	 * 
	 * \param soundLevel_h The sound level in homogeneous conditions
	 * 
	 * \param soundLevel_f The sound level in downward-refraction conditions
	 * 
	 * \param favourableProbability The average occurrence of downward-refraction conditions in the direction of the path
	 * 
	 * \returns The calculated sound level
	 * 
	 * \throws ERRProbability
	 */
	double SoundLevelForPath(double soundLevel_h, double soundLevel_f, double favourableProbability)
	{
		if(favourableProbability > 1)
		{
			printf (".ERROR: probability %4.2f > 1 \n", favourableProbability) ;
			throw new int( ERRProbability );
		}
		double soundLevel = 10 * log10(favourableProbability * pow(10.0, (double)(soundLevel_f/10.0)) + (1 - favourableProbability) * pow(10., (double)soundLevel_h/10.0));
		return soundLevel;
	}

	/*!
	* \brief Calculates the favorable conditions probability for the (SR) direction
	* 
	* \param sourcePos The source position 3D
	* 
	* \param receiverPos The receiver position 3D
	* 
	* \param nbAngles The number of angle probabilities (should be 18)
	* 
	* \param fcpAngles The favorable conditions probabilities for the angles 20, 40, 60, ..., 360
	* 
	* \param angleNorth The north direction (Ox,ON)
	* 
	* \returns the favorable conditions probability for the (SR) direction
	* 
	* \throws ERRAngle
	*/
	double GetFavorableConditionProbability(Position3D const* sourcePos, Position3D const* receiverPos, int nbAngles, double const* fcpAngles, double angleNorth)
	{
		// angle of source-receiver line relative to X-axis, positive = counter clockwise
		double alphaS = 0;
		if (sourcePos->x == receiverPos->x && sourcePos->y == receiverPos->y)
		{
			alphaS = 0;
		}
		else
		{
			alphaS = (180 / PI) * atan2(sourcePos->y - receiverPos->y, sourcePos->x - receiverPos->x);
		}

		// angle relative to north, positive = clockwise
		double alphaNS = angleNorth - alphaS;
		while (alphaNS <= 10)
		{
			alphaNS += 360;
		}
		while (alphaNS > 370)
		{
			alphaNS -= 360;
		}

		// index into array of frequencies of occurrence

		assert (alphaNS > 10) ;
		assert (alphaNS <= 370) ;
		double indexR = (alphaNS - 10) / 20 ;
		assert (indexR > 0) ;
		assert (indexR <= 18) ;
		int indexA = modified_floor(indexR)  ;

		if (indexA > 0 && indexA <= nbAngles)
		{
			// convert probability ([0,100] --> [0,1]) :
			// note that the first value in the table has index 1, i.e. angular sector ]10,30]

			double fp = fcpAngles[indexA - 1] / 100.0;
			return fp;
		}
		else
		{
			printf (".ERROR: angle %4.1f not in array containing favorable conditions probabilities for the angles \n", alphaNS) ;
			throw new int( ERRAngle );
		}
	}

	/*!
	* \brief Calculates the Long-term sound level Leq due to one source at point R, in each given frequency band.
	 * 
	 * \param nbFreq The frequencies number (user data)
	 * 
	 * \param Lw The sound power levels of the source, in each frequency band (user data)
	 * 
	 * \param attH Attenuations due to the propagation between source and receiver in homogeneous conditions, in each frequency band (user data)
	 * 
	 * \param attF Attenuations due to the propagation between source and receiver in downward-refraction conditions, in each frequency band (user data)
	 * 
	 * \param fcp Probability of occurrence of downward-refraction conditions over a long-term period in a given direction, p in [0, 1] (user data)
	 * 
	 * \param LeqH Sound levels due to source Si at point R in homogeneous conditions, in each frequency band (calculated in this function)
	 * 
	 * \param LeqF Sound levels due to source Si at point R in downward-refraction conditions, in each frequency band (calculated in this function)
	 * 
	 * \param LeqLT Long-term sound levels due to source Si at point R, in each given frequency band (calculated in this function)
	 */
	void CalculateLeqLT(int nbFreq, double const* Lw, double const* attH, double const* attF, double fcp, double* LeqH, double* LeqF, double* LeqLT)
	{
		for (int i = 0 ; i < nbFreq ; i++)
		{
			//LAi,F p.16  5.2.1 formula (3)
			LeqH[i] = Lw[i] - attH[i];
			//LAi,H p.17  5.2.2 formula (5)
			LeqF[i] = Lw[i] - attF[i];

			//Li,LT
			LeqLT[i] = SoundLevelForPath(LeqH[i], LeqF[i], fcp);
		}
	}


	// ------------------------------ Attenuation Class -----------------------------------------

	/*!
	* \brief Initialization of Attenuation class with the path and the frequency 
	* 
	* \param path The path where the attenuation will be calculated
	* 
	* \param frequency The frequency to calculation sound attenuation
	*/
	Attenuation::Attenuation(PropagationPath* path, double frequency)
	{
		_path = path; 
		_frequency = frequency; 
		_favourableProbability = 0;
		_displayConsoleStep = path->GetOption(TRACE_DETAILS);
		_withCh = !path->GetOption (FORCE_CH_EQUAL_ONE) ;
		ClearResults();

		// fill frequency map with default values :
		FillFrequencyMap();
	}

	/*!
	* \brief Initialization of Attenuation class with the path, the frequency and the favorable probability
	* 
	* \param path The path where the attenuation will be calculated
	* 
	* \param frequency The frequency to calculation sound attenuation
	* 
	* \param favourableProbability The favorable probability
	*/
	Attenuation::Attenuation(PropagationPath* path, double frequency, double favourableProbability)
	{
		_path = path; 
		_frequency = frequency; 
		_favourableProbability = favourableProbability;
		_displayConsoleStep = path->GetOption(TRACE_DETAILS);
		_withCh = !path->GetOption (FORCE_CH_EQUAL_ONE) ;

		ClearResults();

		// fill frequency map with default values :
		FillFrequencyMap();
	}

	/*!
	 * \brief Attenuation calculation following geometric divergence
	 *
	 * Method for attenuation calculation following geometric divergence
	 *   9.1, formula (20), p. 37
	 * 
	 * \param dist : Direct distance between source and receiver.
	 * 
	 * \returns the calculated attenuation.
	 */
	double Attenuation::DivergenceAttenuationCalculation(double dist)
	{
		double aDiv = 20 * log10(dist) + 11;

		_divergenceAttenuation = aDiv;
		return aDiv;
	}

	/*!
	* \brief Fills the _attenuationCoeffByFrequencyMap map with values of the third octave bands
	*
	* Values taken in the Array 2, p. 37 
	*/
	void Attenuation::FillFrequencyMap()
	{
		_attenuationCoeffByFrequencyMap.clear();
		int frequencyNumber = 24;
		// Nominal median frequencies array
		int NominalMedianFrequency[] = {50, 63, 80, 100, 125, 160, 200, 250, 315, 400, 500, 630, 800, 1000, 1250, 1600, 2000, 2500, 3150, 4000, 5000, 6300, 8000, 10000};
		// Atmospheric attenuation coefficients array
		double AtmosphericAttenuationCoeff[] = {0.07, 0.1, 0.17, 0.25, 0.38, 0.57, 0.82, 1.13, 1.51, 1.92, 2.36, 2.84, 3.38, 4.08, 5.05, 6.51, 8.75, 12.2, 17.7, 26.4, 39.9, 61.1, 93.7, 143.5};
		for(int i = 0; i < frequencyNumber ; i++)
		{
			_attenuationCoeffByFrequencyMap[NominalMedianFrequency[i]] = AtmosphericAttenuationCoeff[i];
		}
	}

	/*!
	* \brief Fills the _attenuationCoeffByFrequencyMap map with values of another map
	* 
	* \param newMap The map to get values
	*/
	void Attenuation::FillFrequencyMap(map<int,double> newMap)
	{
		_attenuationCoeffByFrequencyMap.clear();
		for(map<int,double>::iterator it = newMap.begin(); it != newMap.end(); ++it)
		{    
			_attenuationCoeffByFrequencyMap[it->first] = it->second;
		}



	}



	/*!
	 * \brief Gets attenuation coefficient by frequency
	 *
	 * Gets atmospheric attenuation coefficient for the given frequency
	 * 
	 * \param freq : Nominal median frequency
	 * 
	 * \returns the attenuation coefficient
	 * 
	 * \throws ERRAttCoeffFrequency
	 */
	double Attenuation::AttenuationCoeffByFrequency(double freq)
	{
		double alpha = 0;
		if (_attenuationCoeffByFrequencyMap.find(static_cast<int>(freq)) != _attenuationCoeffByFrequencyMap.end())
		{
			alpha = _attenuationCoeffByFrequencyMap[static_cast<int>(freq)];
		}
		else
		{
			printf (".ERROR: frequency %4.1f not found to get attenuation coefficient \n", freq) ;
			throw new int( ERRAttCoeffFrequency );
		}

		return alpha;
	}

	/*!
	 * \brief Atmospheric absorption
	 * 
	 * Method to calculate atmospheric absorption attenuation 
	 *  9.2, formula (21), p. 37
	 *
	 * \param dist : direct distance between source and receptor
	 * 
	 * \param freq : octave bands center frequency
	 * 
	 * \returns the calculated absorption attenuation.
	 */
	double Attenuation::AtmosphericAbsorptionCalculation(double dist, double freq)
	{
		// alpha : atmospheric attenuation coefficient
		double alpha = AttenuationCoeffByFrequency(freq);

		double aAtm = alpha * dist / 1000.0;
		_atmosphericAbsorption = aAtm;
		return aAtm;
	}


	/*!
	* \brief Calculates ground attenuation, in homogeneous and favorable conditions
	* 
	* Method to calculate ground attenuation, in homogeneous and favorable conditions
	*  9.3 p.38-44
	*
	*/
	void Attenuation::GroundAttenuationCalculation()
	{
		int nbPoints = (int) _path->pathPoints.size();
		if (nbPoints >= 2)
		{
			MeanPlane meanPlane(_path->pathPoints);

			GroundEffect groundEffect(meanPlane.Get_dp(), meanPlane.Get_zEqSource(), meanPlane.Get_zEqReceiver(), meanPlane.Get_Gpath(), meanPlane.Get_Gsource(), _frequency, Asol);
			_groundAttenuation_h = groundEffect.AttenuationCalculationH();
			_groundAttenuation_f = groundEffect.AttenuationCalculationF();

			if (_displayConsoleStep == 1)
			{
				printf(" - AttsolH = %5.1f dB, AttsolF = %5.1f dB \n", _groundAttenuation_h, _groundAttenuation_f);
				printf("  Mean plane :  z = %.3f d + %.3f ; dp = %6.2f \n", meanPlane.Get_aCoeff(), meanPlane.Get_bCoeff(), meanPlane.Get_dp());
				printf("    h : zEqS   = %6.2f, zEqR   = %6.2f \n", groundEffect.Get_zEqSource(), groundEffect.Get_zEqReceiver());
				printf("    f : zEqS_f = %6.2f, zEqR_f = %6.2f \n", groundEffect.Get_zEqSource_f(), groundEffect.Get_zEqReceiver_f());
				printf("  Gpath = %6.2f, G'path = %6.2f \n", meanPlane.Get_Gpath(), groundEffect.get_EquivalentGpath());
			}
		}		
	}

	/*!
	 * \brief Calculates embankment attenuation
	 *
	 *  9.3.5 p.43-44
	 *
	 */
	void Attenuation::EmbankmentAttenuationCalculation()
	{
		_embankmentAttenuation_h = 0;
		_embankmentAttenuation_f = 0;
		if(_path->GetOption(CHECK_EMBANKMENT))
		{
			int nbPoints = (int) _path->pathPoints.size();
			if (nbPoints >= 3)
			{
				Embankment embankment(_path->pathPoints, _frequency);
				double embankmentAttenuation = embankment.GetEmbankmentAttenuation();
				_embankmentAttenuation_h = embankmentAttenuation;
				_embankmentAttenuation_f = embankmentAttenuation;

				if (_displayConsoleStep == 1)
				{
					printf(" - Attenuation talus = %5.1f dB \n", embankmentAttenuation);
				}
			}			
		}
	}

	/*!
	* \brief Calculates border attenuation, in homogeneous and favorable conditions
	* 
	* Method to calculate border attenuation, in homogeneous and favorable conditions
	*  5.2.1 and 5.2.2, p.17-18
	* 
	* \param nbSideDiffractions The number of side diffractions in the path
	* 
	* \throws ERRNoPoint, ERROnePoint, ERRAttCoeffFrequency
	*
	*/
	void Attenuation::BoundaryAttenuationCalculation(int nbSideDiffractions)
	{
		int nbPoints = (int) _path->pathPoints.size();
		if (nbPoints < 1)
		{
			printf (".ERROR: path empty \n") ;
			throw new int( ERRNoPoint );
		}
		else if (nbPoints < 2)
		{
			printf (".ERROR: only one point in the path \n") ;
			throw new int( ERROnePoint );
		}

		// Calculates diffraction or reflection if necessary:
		if (nbPoints >= 3)
		{		
			int fPos = _path->GetFrequencyPosition(_frequency); 
			if(fPos < 0)
			{
				printf (".ERROR: frequency %4.1f not found in path spectrum \n", _frequency) ;
				throw new int( ERRAttCoeffFrequency );
			}

			if(nbSideDiffractions > 0)
			{
				// Side diffraction case
				// no embankment
				_embankmentAttenuation_h = 0;
				_embankmentAttenuation_f = 0;
				// ground attenuation is calculated here
				GroundAttenuationCalculation();

				// Side diffractions and reflections :
				Diffraction diffractionAttenuation(_path->pathPoints, _frequency, fPos, nbSideDiffractions);
				_diffractionAttenuation_h = diffractionAttenuation.Get_DeltaDifSR_h() + diffractionAttenuation.Get_absorptionAttenuation();
				_diffractionAttenuation_f = diffractionAttenuation.Get_DeltaDifSR_h() + diffractionAttenuation.Get_absorptionAttenuation();
			}
			else
			{
				// No side diffraction 
				// ground attenuation calculated in diffraction attenuation, so equals 0
				_groundAttenuation_h = 0;
				_groundAttenuation_f = 0;	

				// embankment attenuation calculated in diffraction attenuation, so equals 0
				_embankmentAttenuation_h = 0;
				_embankmentAttenuation_f = 0;

				// diffraction attenuation :
				Diffraction diffractionAttenuation(_path->pathPoints, _frequency, fPos, _path->GetOption(CHECK_EMBANKMENT));
				if (diffractionAttenuation.Get_pathDifferenceSR_h() == 0 && diffractionAttenuation.Get_pathDifferenceSR_f() == 0) 
				{
					// if diffraction is negligible, ground attenuation is calculated :
					_diffractionAttenuation_h = 0;
					_diffractionAttenuation_f = 0;
					GroundAttenuationCalculation();
					EmbankmentAttenuationCalculation();
				}
				else
				{
					diffractionAttenuation.CalculAttenuationDiffraction (_withCh);
					_diffractionAttenuation_h = diffractionAttenuation.Get_diffractionAttenuation_h();
					_diffractionAttenuation_f = diffractionAttenuation.Get_diffractionAttenuation_f();

					if (_displayConsoleStep == 1)
					{
						printf(" - low height correction Ch : %s \n", _withCh ? "ENABLED" : "DISABLED (Ch = 1)") ;
						printf(" - AttDiffH = %5.1f dB, avec pathDifference = %7.3f, DeltaDiff = %5.1f, ", _diffractionAttenuation_h, diffractionAttenuation.Get_pathDifferenceSR_h(), diffractionAttenuation.Get_DeltaDifSR_h());
						printf("DeltaSol(S,O) = %5.1f, DeltaSol(O,R) = %5.1f, Asol(S,O) = %5.1f", diffractionAttenuation.Get_DeltaSolSO_h(), diffractionAttenuation.Get_DeltaSolOR_h(), diffractionAttenuation.Get_AttsolSO_h());
						if(_path->GetOption(CHECK_EMBANKMENT))
						{
							printf(", DeltaTalus(S,O) = %5.1f ", diffractionAttenuation.Get_DeltaTalusSO_h());
						}
						printf("\n");
						printf(" - AttDiffF = %5.1f dB, avec pathDifference = %7.3f, DeltaDiff = %5.1f, ", _diffractionAttenuation_f, diffractionAttenuation.Get_pathDifferenceSR_f(), diffractionAttenuation.Get_DeltaDifSR_f());
						printf("DeltaSol(S,O) = %5.1f, DeltaSol(O,R) = %5.1f, Asol(S,O) = %5.1f", diffractionAttenuation.Get_DeltaSolSO_f(), diffractionAttenuation.Get_DeltaSolOR_f(), diffractionAttenuation.Get_AttsolSO_f());
						if(_path->GetOption(CHECK_EMBANKMENT))
						{
							printf(", DeltaTalus(S,O) = %5.1f ", diffractionAttenuation.Get_DeltaTalusSO_f());
						}
						printf("\n");
						
					}

					// if diffraction is negligible for homogeneous or favorable path, ground attenuation is calculated :
					if (_diffractionAttenuation_h == 0 || _diffractionAttenuation_f == 0)
					{
						GroundAttenuationCalculation();
						EmbankmentAttenuationCalculation();
						if (_diffractionAttenuation_h != 0)
						{
							_groundAttenuation_h = 0;
							_embankmentAttenuation_h = 0;
						}

						if (_diffractionAttenuation_f != 0)
						{
							_groundAttenuation_f = 0;
							_embankmentAttenuation_f = 0;
						}
					}
				}

				// Add absorption attenuation due to reflections
				_diffractionAttenuation_h += diffractionAttenuation.Get_absorptionAttenuation();
				_diffractionAttenuation_f += diffractionAttenuation.Get_absorptionAttenuation();
			}
		}
		else
		{
			// No diffractions or reflections 
			_diffractionAttenuation_h = 0;
			_diffractionAttenuation_f = 0;

			// ground attenuation calculation :
			GroundAttenuationCalculation();

			// embankment attenuation :
			EmbankmentAttenuationCalculation();
		}	

		// border attenuation :
		_boundaryAttenuation_h = _diffractionAttenuation_h + _groundAttenuation_h + _embankmentAttenuation_h;
		_boundaryAttenuation_f = _diffractionAttenuation_f + _groundAttenuation_f + _embankmentAttenuation_f;

	}

	/*!
	* \brief Calculates total attenuation along the propagation path, in homogeneous and favorable conditions
	* 
	* Method to calculate total attenuation along the propagation path, in homogeneous and favorable conditions
	*  5.2.1, formula (4), p.17 and  5.2.2, formula (6), p.17
	* 
	* \throws ERRSideDiff
	*/
	void Attenuation::AttenuationCalculation()
	{
		if (_displayConsoleStep == 1)
		{
			printf("Attenuation dans la bande d'octave %.1f : \n", _frequency);	
		}

		// Check SR < 2000 :
		if (_path->distSR > 2000)
		{
			printf ("Erreur: la distance de propagation (%.0fm) dpasse la limite de 2000m\n", _path->distSR) ;
			throw new int( ERRMaxDistance );
		}

		// Check side diffractions :
		int nbSideDiffractions = 0;
		for ( int i = 0 ; i < (int) _path->pathPoints.size() ; i++)
		{
			ProfilePointNMPB* profilePoint = _path->pathPoints[i];
			if (profilePoint->ext != NULE && profilePoint->ext->type == ETSideDiffraction_NMPB)
			{
				nbSideDiffractions ++;
			}
		}

		if(nbSideDiffractions > 2)
		{
			printf("Erreur : %d diffractions laterales detectees ; ce nombre ne doit pas depasser 2.\n", nbSideDiffractions);	
			throw new int( ERRSideDiff );
		}

		double divergenceAttenuation = 0;
		if(!_path->GetOption(EXCLUDE_ADIV))
		{
			if(nbSideDiffractions != 0)
			{
				// SR distance (or S'R' if reflections), go through reflections but no side diffractions
				double accuGroundDistance = 0;
				ProfilePointNMPB* precItem = _path->GetSource();
				int nbPoints = (int) _path->pathPoints.size();
				for (int i = 1 ; i < nbPoints ; i++)
				{
					ProfilePointNMPB* curItem = _path->pathPoints[i];
					if (curItem->ext == NULE || curItem->ext->type != ETSideDiffraction_NMPB)
					{
						accuGroundDistance += GroundDistance(&precItem->position3D, &curItem->position3D);
						precItem = _path->pathPoints[i];
					}
				}

				Position2D receivPos;
				Position2D recepPos = _path->GetReceiver()->position2D_withHeight();
				receivPos.d = accuGroundDistance;
				receivPos.z = recepPos.z;
				
				Position2D sourcePos = _path->GetSource()->position2D_withHeight();
				double pathDirectLength = distance2D(&sourcePos, &receivPos);
				
				divergenceAttenuation = DivergenceAttenuationCalculation(pathDirectLength);
			}
			else
			{
				divergenceAttenuation = DivergenceAttenuationCalculation(_path->distSR);
			}
		}

		double atmosphericAbsorption = 0;
		if(!_path->GetOption(EXCLUDE_AATM))
		{
			atmosphericAbsorption = AtmosphericAbsorptionCalculation(_path->distSR, _frequency);
		}

		BoundaryAttenuationCalculation(nbSideDiffractions);
		_attenuation_h = divergenceAttenuation + atmosphericAbsorption + _boundaryAttenuation_h;
		_attenuation_f = divergenceAttenuation + atmosphericAbsorption + _boundaryAttenuation_f;

		if (_displayConsoleStep == 1)
		{					
			printf(" - Aatm = %5.1f dB, Adiv = %5.1f dB \n", atmosphericAbsorption, divergenceAttenuation);			
			printf(" - Attenuations totales : AttH = %5.1f dB, AttF = %5.1f dB \n", _attenuation_h, _attenuation_f);			
			printf(" \n");	
		}
	}

	/*!
	 * \brief Clear calculations results
	 */
	void Attenuation::ClearResults()
	{
		_attenuation_h = 0;
		_attenuation_f = 0;
		_divergenceAttenuation = 0;
		_atmosphericAbsorption = 0;
		_boundaryAttenuation_h = 0;
		_boundaryAttenuation_f = 0;
		_groundAttenuation_h = 0;
		_groundAttenuation_f = 0;
		_diffractionAttenuation_h = 0;
		_diffractionAttenuation_f = 0;
		_embankmentAttenuation_h = 0;
		_embankmentAttenuation_f = 0;
	}

}

