#include "../test_mem/safe_new.h"
#include "RailwayDatabase.h"
#include <math.h>

RailwayDatabase _shared_db ;

// --------------------------------------------------------------------------------------
// standard C interface to railway database functions
// --------------------------------------------------------------------------------------

int NMPB08_LoadRailwayDatabase (const char* filename, bool trace)
{
	 return _shared_db.LoadFile (filename, trace) ;
}

int NMPB08_DumpRailwayDatabase (const char* filename)
{
	return _shared_db.Dump(filename) ;
}

int NMPB08_EnumRailwayDatabase (EnumRailwayEntities enumProc, 
					  		    unsigned int include_types,
							    void* userdata) 
{
	return _shared_db.EnumEntities (enumProc, include_types, userdata) ;
}

int NMPB08_EnumRailwaySources (EnumRailwaySources enumProc,
							   const char* name,
							   void* userdata) 
{
	std::string getID (const char *s) ;
	return _shared_db.EnumSources (enumProc, getID(name), userdata, 0.0) ;
}

// --------------------------------------------------------------------------------------

#define Printf  if (trace_debug) printf

// --------------------------------------------------------------------------------------

static inline double POW10 (double x)
{
	return (x < -99) ? 0.0 : pow(10., x/10.) ;
}

static inline double LOG10 (double x)
{
	return (x <= 0) ? -99.9 : 10 * log10(x) ;
}

template <class T> inline T MIN (T const& x, T const& y)
{
	return (x < y) ? x : y ;
}

template <class T> inline T MAX (T const& x, T const& y)
{
	return (x > y) ? x : y ;
}

// --------------------------------------------------------------------------------------

std::string getID (const char *s)
{
	std::string id ;
	if (s == 0) return id ;

	while (*s != 0)	
	{
		if (*s > 32) id.push_back(*s) ; 
		s++ ;
	}
	return id ;
}

// --------------------------------------------------------------------------------------

static bool check_tag (XMLNode* node, const char* tag)
{
	if (node == 0)
	{
		return false ;
	}
	else if (strcmp (node->GetName(), tag) != 0)
	{
		return false ;
	}
	return true ;
}

// --------------------------------------------------------------------------------------
/*
	read a vector of double values
	input may be coded as text or as separated child items
	names of items and optional attributes are ignored
*/

static bool is_whitespace (const char *s)
{
	return (*s != 0 && *s <= ' ') ;
}

static bool is_separator (const char *s)
{
	return *s != 0 && (*s <= ' ' || *s == ',' || *s == ';') ;
}

static bool get_values (XMLNode* node, Spectrum& data) 
{
	assert (data.size() == 18) ;

	const char*s = node->GetText() ;
	if (s != 0)
	{
		while (is_whitespace(s)) s++ ;
	}

	if (s != 0 && strlen(s) > 0)
	{
		for (int unsigned i = 0 ; i < 18 ; i++)
		{
			if (*s == 0) return false ;
			data[i] = strtod (s, (char**) &s) ;
			while (is_separator(s)) s++ ;
		}
	}
	else
	{
		node = node->GetFirstChild() ;
		for (int unsigned i = 0 ; i < 18 ; i++)
		{
			s = node->GetText() ;
			if (*s == 0) return false ;
			data[i] = strtod (s, (char**) &s) ;
			node = node->GetNextEntity();
		}
	}
	return true ;
}

// --------------------------------------------------------------------------------------

int RailwayDatabase::LoadFile (const char* fileName, bool _trace) 
{
	trace_debug  = _trace ;
	error_code  = 0 ;

	Printf ("Parse file <%s> \n", fileName) ;

	XMLFileLoader cfg ;
	bool ok = cfg.ParseFile (fileName) ;
	if (!ok) 
	{
		Printf (".ERROR:%s is not a valid XML file \n", fileName) ;
		return (error_code = ERROR_XML_PARSER) ;
	}

	XMLNode* root = cfg.GetRoot() ;
	if (! check_tag (root, "train_data"))
	{
		Printf (".ERROR: %s is not a valid configuration file \n", fileName) ;
		return (error_code = ERROR_XML_SCHEMA) ;
	}

	XMLNode* child = root->GetFirstChild() ;
	while (child)
	{
		Printf ("Parse section <%s> \n", child->GetName()) ;

		if (check_tag (child, "unit"))
		{
			ParseUnit (child) ;
		}
		else if (check_tag (child, "train"))
		{
			ParseTrain (child) ;
		}
		else if (check_tag (child, "source"))
		{
			ParseSource (child) ;
		}
		else
		{
			Printf (".WARNING: tag <%s> ignored \n", child->GetName()) ;
		}
		child = child->GetNextEntity() ;
	}

	UnitMap::iterator unit ;
	for (unit = unit_map.begin() ; unit != unit_map.end() ; ++unit)
	{
		for (unsigned int i = 0 ; i < unit->second.source.size() ; i++)
		{
			SourceMap::iterator source = source_map.find (unit->second.source[i].ref) ;
			if (source == source_map.end())
			{
				Printf ("error : unit (%s) uses undefined source (%s) \n",
						unit->first.c_str(), unit->second.source[i].ref.c_str()) ;
			}
		}
	}

	TrainMap::iterator train ;
	for (train = train_map.begin() ; train != train_map.end() ; ++train)
	{
		for (unsigned int i = 0 ; i < train->second.unit.size() ; i++)
		{
			UnitMap::iterator unit = unit_map.find (train->second.unit[i].ref) ;
			if (unit == unit_map.end())
			{
				Printf ("error : train (%s) uses undefined unit (%s) \n",
						train->first.c_str(), train->second.unit[i].ref.c_str()) ;
			}
		}
	}

	return error_code ;
}

// --------------------------------------------------------------------------------------

RailwaySource const* RailwayDatabase::getSource (IDREF ref)
{
	SourceMap::const_iterator source = source_map.find (ref) ;
	if (source == source_map.end()) return 0 ;
	return &source->second ;
}

// --------------------------------------------------------------------------------------

RailwayUnit const* RailwayDatabase::getUnit (IDREF ref)
{
	UnitMap::const_iterator unit = unit_map.find (ref) ;
	if (unit == unit_map.end()) return 0 ;
	return &unit->second ;
}

// --------------------------------------------------------------------------------------

RailwayTrain const* RailwayDatabase::getTrain (IDREF ref)
{
	TrainMap::const_iterator train = train_map.find (ref) ;
	if (train == train_map.end()) return 0 ;
	return &train->second ;
}

// --------------------------------------------------------------------------------------

bool RailwayDatabase::ParseUnit (XMLNode* item) 
{
	IDREF id = getID (item->GetAttribute("id")) ;
	if (id.length() == 0) 
	{
		Printf (".WARNING: missing ID, <unit> ignored \n") ;
		return false ;
	}

	RailwayUnit unit ;

	unit.cref = 1 ;
	unit.length = 20 ;
	unit.vmax = 100 ;

	unit.part_of_train = getID(item->GetAttribute("part")) == "true" ;
	
	XMLNode* child = item->GetFirstChild() ;
	while (child)
	{
		if (check_tag (child, "name"))
		{
			const char* s = child->GetText() ;
			unit.name = std::string(s) ;
		}
		else if (check_tag (child, "length"))
		{
			unit.length = atof (child->GetText()) ;
		}
		else if (check_tag (child, "image"))
		{
			unit.image = getID (child->GetAttribute("file")) ;
		}
		else if (check_tag (child, "vmax"))
		{
			unit.vmax = atof (child->GetText()) ;
		}
		else if (check_tag (child, "refl"))
		{
			unit.cref = atof (child->GetText()) ;
		}
		else if (check_tag (child, "sources"))
		{
			XMLNode* item = child->GetFirstChild() ;
			while (item)
			{
				if (check_tag (item, "source"))
				{
					RailwaySourcePosition pos ;
					if (ParseSourcePosition (item, pos))
					{
						unit.source.push_back (pos) ;
					}
				}
				else
				{
					Printf (".WARNING: tag <%s> ignored \n", item->GetName()) ;
				}
				item = item->GetNextEntity() ;
			}
		}
		else
		{
			Printf (".WARNING: tag <%s> ignored \n", child->GetName()) ;
		}
		child = child->GetNextEntity() ;
	}

	unit_map[id] = unit ;

	return 0 ;
}

// --------------------------------------------------------------------------------------

bool RailwayDatabase::ParseTrain (XMLNode* item) 
{
	IDREF id = getID (item->GetAttribute("id")) ;
	if (id.length() == 0) 
	{
		Printf (".WARNING: missing ID, <source> ignored \n") ;
		return false ;
	}

	RailwayTrain train ;

	XMLNode* child = item->GetFirstChild() ;
	while (child)
	{
		if (check_tag (child, "name"))
		{
			train.name = child->GetText() ;
		}
		else if (check_tag (child, "image"))
		{
			train.image = getID (child->GetAttribute("file")) ;
		}
		else if (check_tag (child, "units"))
		{
			XMLNode* item = child->GetFirstChild() ;
			while (item)
			{
				if (check_tag (item, "unit"))
				{
					RailwayTrainUnit unit ;
					if (ParseTrainUnit(item, unit))
					{
						train.unit.push_back (unit) ;
					}
				}
				else
				{
					Printf (".WARNING: tag <%s> ignored \n", item->GetName()) ;
				}
				item = item->GetNextEntity() ;
			}
		}
		else
		{
			Printf (".WARNING: tag <%s> ignored \n", child->GetName()) ;
		}
		child = child->GetNextEntity() ;
	}

	train_map[id] = train ;

	return true ;
}

// --------------------------------------------------------------------------------------

bool RailwayDatabase::ParseTrainUnit (XMLNode* child, RailwayTrainUnit& unit)
{
	unit.ref = getID(child->GetAttribute("ref")) ;
	if (unit.ref.length() == 0) return false ;

	const char* attr ;
	
	attr = child->GetAttribute ("nb") ;
	if (attr) 
		unit.nb = atoi (attr) ;
	else
		unit.nb = 0 ;

	return (unit.nb > 0) ;
}

// --------------------------------------------------------------------------------------

bool RailwayDatabase::ParseSource (XMLNode* item)
{
	IDREF id = getID (item->GetAttribute("id")) ;
	if (id.length() == 0) 
	{
		Printf (".WARNING: missing ID, <source> ignored \n") ;
		return false ;
	}

	RailwaySource source ;
	assert (source.Lw.size() == 18) ;
	
	XMLNode* child = item->GetFirstChild() ;
	while (child)
	{
		if (check_tag (child, "height"))
		{
			source.height = atof (child->GetText()) ;
		}
		else if (check_tag (child, "vref"))
		{
			source.vref = atof (child->GetText()) ;
		}
		else if (check_tag (child, "slope"))
		{
			source.slope = atof (child->GetText()) ;
		}
		else if (check_tag (child, "hdir"))
		{
			IDREF model = getID (child->GetAttribute("model")) ;
			source.hdir_model = (model == "nmpb" ? HDIR_NMPB : HDIR_OMNI) ;
		}
		else if (check_tag (child, "vdir"))
		{
			IDREF model = getID (child->GetAttribute("model")) ;
			source.vdir_model = (model == "nmpb" ? VDIR_NMPB : VDIR_OMNI) ;
		}
		else if (check_tag (child, "Lw"))
		{
			assert (source.Lw.size() == 18) ;
			get_values (child, source.Lw) ;
		}
		else
		{
			Printf (".WARNING: tag <%s> ignored \n", child->GetName()) ;
		}
		child = child->GetNextEntity() ;
	}

	source_map[id] = source ;

	return true ;
}

// --------------------------------------------------------------------------------------

bool RailwayDatabase::ParseSourcePosition (XMLNode* child, RailwaySourcePosition& pos)
{
	pos.ref = getID(child->GetAttribute("ref")) ;
	if (pos.ref.length() == 0) return false ;

	const char* attr ;
	
	attr = child->GetAttribute ("nb") ;
	if (attr)
		pos.nb = atoi (attr) ;
	else
		pos.nb = 1 ;

	attr = child->GetAttribute ("pos") ;
	if (attr)
		pos.pos = atof (attr) ;
	else
		pos.pos = 0 ;

	attr = child->GetAttribute ("spacing") ;
	if (attr)
		pos.spacing = atof (attr) ;
	else
		pos.spacing = 0 ;

	return true ;
}

// --------------------------------------------------------------------------------------

int RailwayDatabase::Dump (const char* file)
{
	FILE *fp ;
	if (file == 0 || strlen(file) == 0) 
	{
		fp = stdout ;
	}
	else
	{
		fp = fopen (file, "wt") ;
		if (fp == 0) return ERROR_OPEN_FILE ;
	}

	SourceMap::const_iterator source_iter ;
	for (source_iter = source_map.begin() ; source_iter != source_map.end() ; ++source_iter)
	{
		//RailwaySource const& source = source_iter->second ;
		fprintf (fp, "SOURCE: %s\n", source_iter->first.c_str()) ;
	}

	UnitMap::const_iterator unit_iter ;
	for (unit_iter = unit_map.begin() ; unit_iter != unit_map.end() ; ++unit_iter)
	{
		fprintf (fp, "UNIT: %s\n", unit_iter->first.c_str()) ;
	}

	TrainMap::const_iterator train_iter ;
	for (train_iter = train_map.begin() ; train_iter != train_map.end() ; ++train_iter)
	{
		fprintf (fp, "TRAIN: %s\n", train_iter->first.c_str()) ;
	}

	if (fp != stdout) fclose (fp) ;

	return 0 ;
}

// --------------------------------------------------------------------------------------

int RailwayDatabase::EnumEntities (EnumRailwayEntities enumProc, 
								   unsigned int include_types, 
								   void *userdata)
{
	RailwayEntity info ;
	int n = 0 ;

	if ((include_types & ENTITY_UNIT) != 0)
	{
		UnitMap::const_iterator unit_iter ;
		for (unit_iter = unit_map.begin() ; unit_iter != unit_map.end() ; ++unit_iter)
		{
			RailwayUnit const& unit = unit_iter->second ;
			if (unit.part_of_train && (include_types & ENTITY_PARTIAL) == 0) continue ;
			info.id = unit_iter->first.c_str() ;
			info.name = unit.name.c_str() ;
			info.image = unit.image.c_str() ;
			info.length = unit.length ;
			info.vmax = unit.vmax ;
			info.type = ENTITY_UNIT ;
			if (unit.part_of_train) info.type |= ENTITY_PARTIAL ;
			n++ ;
			if (! enumProc (info, userdata)) return n ;
		}
	}

	if ((include_types & ENTITY_TRAIN) != 0)
	{
		TrainMap::const_iterator train_iter ;
		for (train_iter = train_map.begin() ; train_iter != train_map.end() ; ++train_iter)
		{
			RailwayTrain const& train = train_iter->second ;
			info.id = train_iter->first.c_str() ;
			info.name = train.name.c_str() ;
			info.image = train.image.c_str() ;
			double length = 0 ;
			double vmax = 999 ;
			for (unsigned int i = 0 ; i < train.unit.size() ; i++)
			{
				RailwayUnit const* unit = getUnit (train.unit[i].ref) ;
				if (unit)
				{
					length += unit->length ;
					vmax = MIN (vmax, unit->vmax) ;
				}
			}
			info.length = length ;
			info.vmax = vmax ;
			info.type = ENTITY_TRAIN ;
			n++ ;
			if (! enumProc (info, userdata)) return n ;
		}
	}

	return n ;
}

// --------------------------------------------------------------------------------------

int RailwayDatabase::EnumSources (EnumRailwaySources enumProc,
							      IDREF id,
							      void* userdata,
								  double pos_unit) 
{
	int n = 0 ;
	if (id.size() == 0) return n ;

	TrainMap::const_iterator train_iter ;
	train_iter = train_map.find (id) ;	
	if (train_iter != train_map.end())
	{
		RailwayTrain const& train = train_iter->second ;
		for (unsigned int i = 0 ; i < train.unit.size() ; i++)
		{
			RailwayUnit const* unit = getUnit (train.unit[i].ref) ;
			if (unit)
			{
				for (int j = 0 ; j < train.unit[i].nb ; j++)
				{
					n += EnumSources (enumProc, train.unit[i].ref, userdata, pos_unit) ;
					pos_unit += unit->length ;
				}
			}
		}
		return n ;
	}

	UnitMap::const_iterator unit_iter ;
	unit_iter = unit_map.find (id) ;	
	if (unit_iter != unit_map.end())
	{
		RailwayUnit const& unit = unit_iter->second ;
		for (unsigned int i = 0 ; i < unit.source.size() ; i++)
		{
			double pos	   = pos_unit + unit.source[i].pos ;
			int    nb      = unit.source[i].nb ;
			double spacing = unit.source[i].spacing ;
			RailwaySource const* source = getSource (unit.source[i].ref) ;
			if (source)
			{
				for (int j = 0 ; j < nb ; j++)
				{
					RailwayElementarySource info ;

					info.id     = unit.source[i].ref.c_str() ;
					info.pos    = pos ;
					info.height = source->height ;
					info.hdir_model = source->hdir_model ;
					info.vdir_model = source->vdir_model ;
					info.vref = source->vref ;
					info.slope = source->slope ;
					info.Lw = &source->Lw[0] ;

					if (!enumProc (info, userdata)) return n ;
					
					pos += spacing ;
					n++ ;
				}
			}
		}
	}

	return n ;
}
