#ifndef _RAILWAYEMISSION_INCLUDED_

#include "RailwayEmissionNMPB08.h"
#include "RailwayDatabase.h"
#include "math.h"
#include <vector>

// on Unix-like systems there is no way to hide local classes ant to prevent them from 
// being replaced by other classes with similar names in other libraries. The use of 
// unusual names may help preventing name clashes.

#ifndef _WIN32
#define RailwaySourceModel		_Local_RWEMN8_RailwaySourceModel_
#endif

extern RailwayDatabase _shared_db ;

/*!
 *	\brief the railway source model links traffic data to the sound power 
 *	       radiated from equivalent sources
 */
class RailwaySourceModel
{
public:

	/*!
	 * \brief constructor
	*/
	RailwaySourceModel (void)
	: db(), traffic(), emission(), options(), maxInteractions(), nbInteractions(), difInteraction()
	{
		db = &_shared_db ;
		options = 0 ;
		createEmission() ;
		createTraffic() ;
	}

	/*!
	 * \brief destructor
	*/
	virtual ~RailwaySourceModel (void)
	{
		deleteEmission() ;
		deleteTraffic() ;
	}

	/*!
	 * \brief initialise the traffic record
	 *
	 * \arg hours duration of the period of obervation in hours
	*/
	bool CreateTraffic (double hours)
	{
		if (traffic.nbComponents > 0) ClearTraffic() ;
		traffic.nbHours = hours ;
		return true ;
	}

	/*!
	 * \brief clears the traffic flow
	*/
	bool ClearTraffic (void)
	{
		for (unsigned int i = 0 ; i < traffic.nbComponents ; i++) free (traffic.component[i].id) ;
		traffic.nbComponents = 0 ;
		return true ;
	}

	/*!
	 * \brief set the rail correction 
	*/
	bool SetRailCorrection (NMPB_RailCorrectionType corr)
	{
		traffic.railCorrection = corr ;
		return true ;
	}

	/*!
	 * \brief get the rail correction 
	*/
	NMPB_RailCorrectionType GetRailCorrection (void)
	{
		return traffic.railCorrection  ;
	}

	 /*!
	 * \brief add trains or units to the traffic flow
	 *
	 * \arg name	identification of the train or unit as specified in the database
	 *
	 * \arg nb		number of movements over the observation period
	 *
	 * \arg speed	speed of the trains in km/h
	*/
	bool AddTraffic (const char* name, double nb, double speed);

	/*!
	 * \brief  defines the angles of emission for the calculation of apparent sound power
	 *
	 * \arg sin_h	sinus of the horizontal angle, 0 = plane perpendicular to the track
	 *
	 * \arg sin_v	sinus of the vertical angle, 0 = horizontal plane
	*/
	bool SetEmissionAngles (double sin_h, double sin_v) 
	{
		int err = 0 ;
		if (fabs(sin_h) <= 1.0) emission.sin_h = sin_h ; else err++ ;
		if (fabs(sin_v) <= 1.0) emission.sin_v = sin_v ; else err++ ;
		return (err == 0) ;
	} ;

	/*!
	 * \brief exception thrown when an invalid source is detected and 
	 *        \ref OPTION_MODIFY_HEIGHTS is not set  
    */
	class InvalidHeight { } ;

	/*!
	 * \brief calculates and returns the corresponding emission model in terms
	 *        of equivalent source positions and (apparent) sound power
	 *
	 * \return pointer to the emission model, null pointer in case of error
	 *
	 * \note in case the traffic refers trains and units with sources
	 *       defined at heights other than those of the emission model,
	 *       an error willbe detected and the function will return a
	 *       NULL pointer. This behaviour can be avoided by setting the
	 *       \ref OPTION_MODIFY_HEIGHTS option.
	*/
	RailwayEmission const* GetEmission (ScreenBodyInteraction* screenBodyInteraction) ;

	/*!
	 * \brief return the current traffic data associated with the model
	 *
	 * \return pointer to the traffic data, null pointer in case of error
	 *
	*/
	RailwayTraffic const* GetTraffic (void)
	{
		return &traffic ;
	}

	/*!
	 * \brief enables or disables selected options
	 *
	 * \param option options to be set or cleared
	 *
	 * \param on_off if true, options will be set, otherwise options will be cleared
	 *
	 * \return the modified set of options
	 */
	unsigned int SetOptions (unsigned int option, bool on_off)
	{
		if (on_off) 
			options |= option ;
		else
			options &= ~option ;
		return options ;
	}

	/*!
	 * \brief return the currently selected options
	 *
	 * \param option options to be reported
	 *
	 * \return the selected set of options
	 */
	unsigned int GetOptions (unsigned int option)
	{
		return options & option ;
	}

private :

	/*!
	 * \brief create and initialise the emission model, specifically, sets the 
	 *        source heights to the 3 values supported in the NMPB-2008 model
	 */
	void createEmission (void) ;

	/*!
	 * \brief cleanup and deletes emission model
	 */
	void deleteEmission (void) ;

	/*!
	 * \brief creates the traffic model
	 */
	void createTraffic (void) ;

	/*!
	 * \brief deletes the traffic model
	 */
	void deleteTraffic (void) ;

	/*!
	 * \brief prepares the emission model, i.e. resets all source powers to 0 
	 */
	void prepareEmission (void) ;
	/*!
	 * \brief adds all traffic to the emission model
	 */
	void addTraffic (RailwayTraffic const& traffic) ;
	/*!
	 * \brief adds a single train to the emission model
	 */
	void addTrain   (RailwayTrain   const& train,  double speed, double flow) ;
	/*!
	 * \brief adds a single unit to the emission model
	 */
	void addUnit    (RailwayUnit    const& unit,   double speed, double flow) ;
	/*!
	 * \brief adds a single elementary source to the emission model
	 */
	void addSource  (RailwaySource  const& source, double speed, double flow) ;
	/*!
	 * \brief finalizes the emission model, identifies active/inactive sources
	 *        and calculates sound power levels in dB
	 */
	void finaliseEmission (void) ;

	/*!
	 * \brief distribution of sound power over different source heights
	 */
	struct SourceDistribution
	{
		int		index ;		/*!< index of source height in the emission model */
		double  weight ;    /*!< weighting of sound power */
	} ;
	/*!
	 * \brief calculate distribution of sound power over different source heights
	 * 
	 * in case the source height in the database does not match one the heights defined
	 * in the emission model, the coresponding sound power will be distributed over the
	 * nearest source heights. Note that \ref OPTION_MODIFY_HEIGHTS must be set in order to
	 * enable automatic distribution of sound powers over different source heights
	 */
	int getSourceDistribution (RailwaySource const& source, std::vector<SourceDistribution>& distrib) ;

	/*!
	 * \brief calculate the horizontal directivity
	 */
	double getHorizontalDirectivity (int model, double sin_h, double freq) ;
	/*!
	 * \brief calculate the vertical directivity
	 */
	double getVerticalDirectivity (int model, double sin_v, double freq) ;
	
	/*!
	 * \brief calculate attenuation due to diffraction (used in body/barrier interaction)
	 */
	double getDiffraction (double delta, double freq) ;

	/*!
	 * \brief calculate the correction for screen / body interaction as a function of
	 *        source height, number of interactions and frequency, assuming the car
	 *		  body is a perfect reflector (reflection on the car body is taken into
	 *        account separately as part of sound power model for each unit).
	 */
	void prepareScreenBodyCorrection (ScreenBodyInteraction*) ;

	typedef std::vector<std::vector<std::vector<double> > > DifInteraction ;

	RailwayDatabase*	db ;			  /*!< access to shared database */
	RailwayTraffic		traffic ;		  /*!< railway traffic data */
	RailwayEmission		emission ;		  /*!< equivalent emission model */
	unsigned int		options ;		  /*!< options */
	int					maxInteractions ; /*!< maximum nb of car body / barrier interactions */
	int					nbInteractions ;  /*!< current nb of car body / barrier interactions */
	DifInteraction		difInteraction ;  /*!< weighting for car/body interaction */
} ;

#endif

