#ifndef _RAILWAY_EMISSION_NMPB_INCLUDED_
#define _RAILWAY_EMISSION_NMPB_INCLUDED_

// --------------------------------------------------------------------------------------
/*!
* \file RailwayEmissionNMPB08.h
* \brief Emission model for railway sources to be used in the context of the
*        NMPB-2008 prediction model for environmental noise
* \author CSTB
* \version 1.0
*
* \mainpage NMPB08 Railway Emission Model 
  \n 
   The railway emission model is defined in NF S31 133:2011. The standard defines the source
   model in terms of souce heights and directivity but does not include actual
   data for the associated sound powers.
  \n\n
   The sound power calculations in this library are based on a XML-based database format
   defined by CSTB, RFF and SNCF.
  \n\n
   Actual data is available from RFF.
  \n\n
  \section main Public functions implemented in the RailwayEmissionNMPB08 library 
  \subsection sub1 Loading and consulting the train database
  - \ref NMPB08_LoadRailwayDatabase
  - \ref NMPB08_DumpRailwayDatabase
  - \ref NMPB08_EnumRailwayDatabase
  \subsection sub2 Defining a railway traffic
  - \ref NMPB08_CreateRailwayTraffic
  - \ref NMPB08_DeleteRailwayTraffic
  - \ref NMPB08_ClearRailwayTraffic
  - \ref NMPB08_AddRailwayTraffic
  \subsection sub3 Getting the equivalent source model
  - \ref NMPB08_SetRailwayEmissionAngles
  - \ref NMPB08_SetRailCorrection
  - \ref NMPB08_GetRailCorrection
  - \ref NMPB08_GetRailwayEmission
  - \ref NMPB08_EnumRailwaySources
  \subsection sub4 Setting and getting options
  - \ref NMPB08_SetRailwayOptions
  - \ref NMPB08_GetRailwayOptions
*/

// --------------------------------------------------------------------------------------

/*!
 * \brief compiler specific options for creating shared libraries
 */

#ifndef _COMPILE_NMPB
	#ifdef _WIN32 
		#define _COMPILE_NMPB extern "C" __declspec(dllexport)
	#else
		#define _COMPILE_NMPB extern "C"
	#endif 
#endif

// --------------------------------------------------------------------------------------
/*!
 * \brief error types returned from the NMPB08_Railway software library
 */
enum NPMP08_RailwayEmission_Error
{
	ERROR_XML_PARSER     = 1, /*!< file could not be openened or parsed */
	ERROR_XML_SCHEMA     = 2, /*!< incompatible XML file */
	ERROR_OPEN_FILE      = 3, /*!< file could not be openened */
	ERROR_INVALID_ID     = 4,  /*!< identifier not defined in database*/
	ERROR_INVALID_HEIGHT = 5   /*!< source height not supported in emission model */
} ;

// --------------------------------------------------------------------------------------
/*!
 * \brief predefined horizontal directivity models
 */
enum NMPB_RailwayEmission_HorizontalDirectivity
{
	HDIR_OMNI = 0, /*!< omnidirectionnel */
	HDIR_NMPB = 1  /*!< horizontal directivity as defined in NF S31-133 */
} ;

// --------------------------------------------------------------------------------------
/*!
 * \brief predefined vertical directivity models
 */
enum NMPB_RailwayEmission_VerticalDirectivity
{
	VDIR_OMNI = 0, /*!< omnidirectional */
	VDIR_NMPB = 1, /*!< vertical directivity as defined in NF S31-133 */
	VDIR_HEMI = 2  /*!< hemispheric emission upwards */
} ;

// --------------------------------------------------------------------------------------
/*!
 * \brief options that modify the behaviour of the calculation model
 *
 * \note OPTION_NO_DIRECTIVITY is useful if the directivity of the railway sources
 *       is integrated as part of the propagation model
 *
 * \note OPTION_MODIFY_HEIGHTS is useful to adjust databases for different 
 *       propagation models, i.e. when source heights in the database do not
 *	     match those of the emission model
*/
enum NMPB_RailwayEmission_Options
{
	OPTION_TRACE_DEBUG		   = 1, /*!< trace intermediate results to stdout */
	OPTION_NO_DIRECTIVITY	   = 2, /*!< do not include directivity */
	OPTION_MODIFY_HEIGHTS	   = 4,  /*!< distribute sound powers over source heights */
	OPTION_DEYGOUT_DIFFRACTION = 8 /*!< use Deygout approximation for body/barrier interaction */
} ;

// --------------------------------------------------------------------------------------
/*!
 * \brief correction for rail and mounting conditions as specified in the document
    \"Production des cartes strategiques des grands axes routiers et ferrovaires\",
	published by SETRA, August 2007.
	\n\n
	Note that the following values are indicative and not integral part of the method.
	\n\n
	- RAIL_LONG_TRAVERSES_BETON  = 0 dB(A)
	- RAIL_LONG_TRAVERSES_AUTRE  = 3 dB(A)
	- RAIL_COURT_TRAVERSES_BETON = 3 dB(A)
	- RAIL_COURT_TRAVERSES_AUTRE = 6 dB(A)
	\n\n
	The end user may supply other values, e.g. for points and metal bridges
*/
typedef double NMPB_RailCorrectionType ;

const NMPB_RailCorrectionType  RAIL_LONG_TRAVERSES_BETON  = 0.0 ;
const NMPB_RailCorrectionType  RAIL_LONG_TRAVERSES_AUTRE  = 3.0 ;
const NMPB_RailCorrectionType  RAIL_COURT_TRAVERSES_BETON = 3.0 ;
const NMPB_RailCorrectionType  RAIL_COURT_TRAVERSES_AUTRE = 6.0 ;
const NMPB_RailCorrectionType  RAIL_ZONE_APPAREILS_VOIE   = 6.0 ;

// --------------------------------------------------------------------------------------
/*!
 * \brief elementary emission data represents apparent sound power
 *        for an equivalent line source emitted in a given direction
*/
struct RailwayEquivalentSource
{
	double	 height ; /*!< height of the equivalent point source */
	bool	 has_Lw ; /*!< true if the equivalent source has a finite sound power */
	double*	 Lw_dir ; /*!< sound power per third octave band */
} ;

// --------------------------------------------------------------------------------------
/*!
 * \brief the emission associated with a railway traffic is represented as a set
 *        of elementary point sources with associated sound power and directivity 

 * \note the horizontal angle is coded as the sinus of the angle between the direction
 *       of propagation and the vertical plane perpendicular to the track
 * \note the vertical angle is coded as the sinus of the angle between the direction
 *       of propagation and the horizontal plane containing the track
*/
struct RailwayEmission
{
	unsigned int				nbFreq ;	/*!< number of frequency bands */
	double*						freq ;		/*!< table of frequency bands */
	double						sin_h ;		/*!< horizontal component of the direction of propagation */
	double						sin_v ;		/*!< vertical component of the direction of propagation */
	unsigned int				nbSources ;	/*!< number of equivalent sources */
	RailwayEquivalentSource*	source ;	/*!< table of equivalent sources */
} ;

// --------------------------------------------------------------------------------------
/*!
 * \brief a single railway traffic component. Each component is defined as a number of 
 *        units (or trains) of a given type that ciruclate at the same speed on the same 
 *        track, the same speed. A complete traffic may be composed of one or more 
 *        components.
*/
struct RailwayTrafficComponent
{
	char*	id ;      /*!< identification of a unit or train */
	double	number ;  /*!< number of units/train per period of reference */
	double	speed ;   /*!< speed in km/h */
} ;

/*!
 * \note the period of reference is defined when the traffic is created, see
 *       \ref NMPB08_CreateRailwayTraffic
 */

// --------------------------------------------------------------------------------------
/*!
	railway traffic represents number of units (or trains) running on
	one track at specified speeds during a period of reference (in hours)
*/
struct RailwayTraffic
{
	double						nbHours ;        /*!< duration of the period of reference in hours */
	NMPB_RailCorrectionType	    railCorrection ; /*!< correction for rail and mounting conditions */
	unsigned int				nbComponents ;   /*!< number of traffic components */
	unsigned int				maxComponents ;  /*!< for internal use only */
	RailwayTrafficComponent*	component ;      /*!< table of traffic components */
} ;

// --------------------------------------------------------------------------------------
/*!
	configuration of interaction between the train body and a nearby screen 
*/
struct ScreenBodyInteraction
{
	int		 maxInteractions ;  /*!< number of interactions to be taken into account */
	double	 dS ;				/*!< horizontal distance from source to screen */
	double   hS ;				/*!< height of screen relative to the head of the track */
	double   dR ;				/*!< horizontal distance from screen to receiver */
	double   hR ;				/*!< height of receiver relative to the head of the track */
	double*  alpha ;			/*!< absoprtion applied to the inner side of the barrier */
} ;

// --------------------------------------------------------------------------------------
/*!
 * \brief loads the train database from an external file
 *
 * \param filename name of the external file
 *
 * \param log_stdout if true, prints messages to stdout
 *
 * \return 0 if successful, otherwise an error code
 */
_COMPILE_NMPB int NMPB08_LoadRailwayDatabase (const char* filename, bool log_stdout = false) ; 

// --------------------------------------------------------------------------------------
/*!
 * \brief writes the contents of the database to a file
 *
 * \param filename the name of the output file 
 *        in case the filename is a zero pointer or a zero length string, the
 *        output will be written to stdout
 *
 * \return 0 if successful, otherwise an error code
 */
_COMPILE_NMPB int NMPB08_DumpRailwayDatabase (const char* filename = 0) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief constants used for enumerating the contents of the database
 */

enum NMPB_RailwayEmission_Entities
{
	ENTITY_UNIT     = 1, /*!< unit */
	ENTITY_TRAIN    = 2, /*!< train*/
	ENTITY_PARTIAL  = 4, /*!< unit, only to be used as part of a train */
	ENTITY_TRAFFIC  = ENTITY_TRAIN | ENTITY_UNIT, /*!< traffic units only */
	ENTITY_ALL      = ENTITY_TRAIN | ENTITY_UNIT | ENTITY_PARTIAL /*!< all units and trains  */
} ;

// --------------------------------------------------------------------------------------
/*!
 * \brief structure used for enumerating the contents of the database
 */
struct RailwayEntity
{
	unsigned int		type ;	  /*!< unit, train or part of a train */
	const char*			id ;	  /*!< identifier */
	const char*			name ;    /*!< full name */
	const char*			image ;   /*!< name of image file */
	double				length ;  /*!< lenght in meters */
	double				vmax ;    /*!< maximum speed in km/h */
} ;

// --------------------------------------------------------------------------------------
/*!
 * \brief user defined callback function for enumerating the contents of the database
 */
typedef bool (*EnumRailwayEntities) (RailwayEntity const& info, void* userdata) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief enumerate the contents of the database
 *
 * \param enumProc user defined callback function
 *
 * \param include_types type of database entries to be enumerated
 *
 * \param userdata user defined data to be passed to the callback function
 */
_COMPILE_NMPB int NMPB08_EnumRailwayDatabase (EnumRailwayEntities enumProc, 
					  						  unsigned int include_types = ENTITY_TRAFFIC,
											  void* userdata = 0) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief create a context for converting railway traffic into an equivalent source model
 *
 * \param nb_hours duration of the period of reference in hours
 *
 * \return a handle to an internal structure to be uses in consecutive calls to other 
 *         function inside the library, a NULL pointer if error occured
 */
_COMPILE_NMPB void* NMPB08_CreateRailwayTraffic (double nb_hours) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief empties the traffic associated with the source model
 *
 * \param id handle to the internal structure used by the traffic model
 *
 * \return O if sucessful, an error code otherwise
 */
_COMPILE_NMPB int NMPB08_ClearRailwayTraffic (void* id) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief adds a traffic component to the source model
 *
 * \param id handle to the internal structure used by the traffic model
 *
 * \param unit_or_train identification of the type of unit or train
 *
 * \param number number of units or trains passing during the reference period
 *
 * \param speed speed in km/h
 *
 * \return O if sucessful, an error code otherwise
 */
_COMPILE_NMPB int NMPB08_AddRailwayTraffic (void* id, const char* unit_or_train, double number, double speed) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief enables or disables selected options
 *
 * \param id handle to the internal structure used by the traffic model
 *
 * \param option options to be set or cleared
 *
 * \param on_off if true, options will be set, otherwise options will be cleared
 *
 * \return the modified set of options
 */
_COMPILE_NMPB unsigned int NMPB08_SetRailwayOptions (void* id, unsigned int option, bool on_off) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief return the currently selected options
 *
 * \param id handle to the internal structure used by the traffic model
 *
 * \param option options to be reported
 *
 * \return the selected set of options
 */
_COMPILE_NMPB unsigned int NMPB08_GetRailwayOptions (void* id, unsigned int option) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief set the rail correction
 *
 * \param id handle to the internal structure used by the traffic model
 *
 * \param corr new correction for the rail / mounting condition
 *
 * \return O if sucessful, an error code otherwise
 */

_COMPILE_NMPB unsigned int NMPB08_SetRailCorrection (void* id, NMPB_RailCorrectionType corr) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief returns the current value of rail correction
 *
 * \param id handle to the internal structure used by the traffic model
 *
 * \return the current value of the rail correction
 */

_COMPILE_NMPB NMPB_RailCorrectionType NMPB08_GetRailCorrection (void* id) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief returns the equivalent source model for the current traffic state
 *
 * \param id handle to the internal structure used by the traffic model
 *
 * \param screenBodyInteraction configuration of interaction between the train body 
		  and a nearby screen One interaction corresponds to a double reflection, once on 
 *		  the barrier, once on the car body. The equivalent reflection coefficient of the 
 *        car bodies is encoded in the database and will modify the equivalent sound
 *		  power of the image sources created through reflection. Absorption on the inner
 *        side of the barrier is also taken into account (see NF S 31-133, section 7.4.6).
 *
 * \return a pointer to the equivalent source model, a null pointer in case an invalid
 *         source height has been detected in the selected trains and units
 *
 * \note if OPTION_H_INTERPOLATE is set, sources heights defined in the database will
 *       be automatically adapted to the 3 sources heights defined in the NMPB model.
 *		 This is achieved by distributing the acoustical power over the predefined
 *       source heights proportional to the difference in height
 */
_COMPILE_NMPB RailwayEmission const* NMPB08_GetRailwayEmission (void* id, ScreenBodyInteraction* screenBodyInteraction = 0) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief sets de emission angles for the equivalent source model
 *
 * \param id handle to the internal structure used by the traffic model
 *
 * \param sin_h sinus of the angle of the propagation direction with the vertical plane
 *              perpendicular to the track
 *
 * \param sin_v sinus of the angle of the propagation direction with the horizontal plane
 *              containing the track
 *
 * \return 0 if successful, an error code otherwise
 *
 * \note if the OPTION_NO_DIRECIVITY is set, this function has no effect on the 
 *       reported equivalent source model 
 */
_COMPILE_NMPB int NMPB08_SetRailwayEmissionAngles (void* id, double sin_h, double sin_v) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief destroys the internal data structure used for converting railway traffic
 *		  data into an equivalent source model.
 *
 * \param id handle to an internal data structure as returned by /ref NMPB08_CreateRailwayTraffic.
 *
 * \return 0 if successful, an error code otherwise.
 */
_COMPILE_NMPB int NMPB08_DeleteRailwayTraffic (void* id) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief structure used for enumerating the equivalent acoustical sources 
 *        associated with a train or a unit
 */
struct RailwayElementarySource
{
	const char*			id ;  /*!< identifier */
	double			   pos ;  /*!< position of the source along the train */
	double			height ;  /*!< height of the equivalent point source */
	double			  vref ;  /*!< reference speed */
	double			 slope ;  /*!< speed dependency of the sound power */
	int		 	hdir_model ;  /*!< horizontal directivity (see \ref NMPB_RailwayEmission_HorizontalDirectivity) */
	int			vdir_model ;  /*!< vertical directivity (see \ref NMPB_RailwayEmission_VerticalDirectivity) */
	double const*	    Lw ;  /*!< sound power spectrum at the reference speed */ 
} ;

// --------------------------------------------------------------------------------------
/*!
 * \brief user defined callback function for enumerating the equivalent acoustical
 *        sources associated with a train or a unit
 */
typedef bool (*EnumRailwaySources) (RailwayElementarySource const& info, void* userdata) ;

// --------------------------------------------------------------------------------------
/*!
 * \brief enumerate the equivalent acoustical sources associated with a train or unit
 *
 * \param enumProc user defined callback function
 *
 * \param name identification a train or unit 
 *
 * \param userdata user defined data to be passed to the callback function
 */
_COMPILE_NMPB int NMPB08_EnumRailwaySources (EnumRailwaySources enumProc,
						  					 const char* name,
											 void* userdata) ;

#endif
