/*!
* \file RoadEmissionNMPB08.cpp
* \brief Definition of the dll functions that can be called by external software
* \author CSTB
* \version 1.0
*/

#include "RoadEmissionNMPB08.h"
#include "CalculRoadEmission.h"
#include <vector>
#include <math.h>
#include <stdio.h>
using namespace std;
using namespace CalculRoadEmissionNMPB;

 /*!
  * \brief Calculation of the Sound power level per source line meter
  *
  * \param roadTraffic The road traffic data
  * 
  * \param spectrumType The spectrum type
  * 
  * \returns The Sound power level per source line meter, for each third-octave band
  */
 double* NMPB08_Lwm (RoadTraffic* roadTraffic, RoadSpectrumType spectrumType)
 {
	 try
	 {
		CalculRoadEmission calculRoadEmission(spectrumType, roadTraffic->surfaceType);
		return calculRoadEmission.SoundPowerLevelPerMeter(roadTraffic);
	 }
	 catch (const std::exception& e)
	 {
		 printf ("ERREUR : %s \n", e.what()) ;
		 return NULL;
	 }
 }

 /*!
 * \brief Calculation of the rolling component of the Sound power level per source line meter
 *
 * \param roadTraffic The road traffic data
 * 
 * \param spectrumType The spectrum type
 * 
 * \returns The rolling component of the Sound power level per source line meter, for each third-octave band
 */
 double* NMPB08_Lwm_rolling (RoadTraffic* roadTraffic, RoadSpectrumType spectrumType)
 {
	 try
	 {
		 CalculRoadEmission calculRoadEmission(spectrumType, roadTraffic->surfaceType);
		 return calculRoadEmission.Lwm_rolling(roadTraffic);
	 }
	 catch (const std::exception& e)
	 {
		 printf ("ERREUR : %s \n", e.what()) ;
		 return NULL;
	 }
 }

 /*!
 * \brief Calculation of the traction component of the Sound power level per source line meter
 *
 * \param roadTraffic The road traffic data
 * 
 * \param spectrumType The spectrum type
 * 
 * \returns the traction component of the Sound power level per source line meter, for each third-octave band
 */
 double* NMPB08_Lwm_traction (RoadTraffic* roadTraffic, RoadSpectrumType spectrumType) 
 {
	 try
	{
		 CalculRoadEmission calculRoadEmission(spectrumType, roadTraffic->surfaceType);
		 return calculRoadEmission.Lwm_traction(roadTraffic);
	 }
	 catch (const std::exception& e)
	 {
		 printf ("ERREUR : %s \n", e.what()) ;
		 return NULL;
	 }
 }

 /*!
  * \brief Internal Utility function, copies return values to static local memory 
  * 
  * \param Lw pointer to dynamically allocated return values
  * 
  * \param specType spectral contents of return values
  * 
  * \returns a pointer to a local (static) copy of the return values
  */
 static double Lw_out[18] ;
 static double* static_copy (double* Lw, RoadSpectrumType specType)
 {
	 int nbFreq = 1 ;
	 switch (specType)
	 {
	 case Spectrum_oct_A :
	 case Spectrum_oct_lin :
		 nbFreq = 6 ;
		 break ;
	 case Spectrum_3oct_A :
	 case Spectrum_3oct_lin :
		 nbFreq = 18 ;
		 break ;
	 }
	 for (int i = 0 ; i < nbFreq ; ++i) Lw_out[i] = Lw[i] ;
	 return Lw_out ;
 }

 /*!
  * \brief Calculation of the rolling component of the Sound power level per source line meter for a vehicle type
  * 
  * \param type The vehicle type
  * 
  * \param vehicleSpeed The vehicle speed
  * 
  * \param surfaceType The surface type
  * 
  * \param surfaceAge The surface age
  * 
  * \param specType The spectrum type
  * 
  * \returns the rolling component of the Sound power level per source line meter, for each third-octave band
  */
 double* NMPB08_Lw_rolling (RoadVehicleType type, 
  						    double vehicleSpeed, 
							RoadSurfaceType surfaceType, 
							double surfaceAge,
							RoadSpectrumType specType) 
 {
	try
	{
		 CalculRoadEmission calculRoadEmission (specType, surfaceType);
		 double* Lw = calculRoadEmission.Lw_rolling (type,vehicleSpeed, surfaceType, surfaceAge);
		 return static_copy (Lw, specType) ;
	 }
	 catch (const std::exception& e)
	 {
		 printf ("ERREUR : %s \n", e.what()) ;
		 return NULL;
	 }
 }

 /*!
  * \brief Calculation of the traction component of the Sound power level per source line meter for a vehicle type
  * 
  * \param type The vehicle type
  * 
  * \param vehicleSpeed The vehicle speed
  * 
  * \param flowType The flow type
  * 
  * \param ramp The ramp
  * 
  * \param specType The spectrum type
  * 
  * \returns the traction component of the Sound power level per source line meter, for each third-octave band
  */
 double* NMPB08_Lw_traction (RoadVehicleType type,
							 double vehicleSpeed,
							 RoadFlowType flowType,
							 double ramp,
							 RoadSpectrumType specType) 
 {
	 try
	 {
		 CalculRoadEmission calculRoadEmission(specType, RoadSurface_Default);
		 double* Lw = calculRoadEmission.Lw_traction(type, vehicleSpeed, flowType, ramp);
		 return static_copy (Lw, specType) ;
	 }
	 catch (const std::exception& e)
	 {
		 printf ("ERREUR : %s \n", e.what()) ;
		 return NULL;
	 }
 }

 /*!
  * \brief Definition of a new road surface type with data needed for rolling component and for spectrum
  * 
  * \param idRoadSurface The road surface id
  * 
  * \param AVL The AVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
  * 
  * \param BVL The BVL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
  * 
  * \param Vref_VL The Vref_VL value for the Lr_w_VL calculation where Lr_w_VL = BVL + AVL * log10(V/Vref_VL)
  * 
  * \param APL The APL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
  * 
  * \param BPL The BPL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
  * 
  * \param Vref_PL The Vref_PL value for the Lr_w_PL calculation where Lr_w_PL = BPL + APL * log10(V/Vref_PL)
  * 
  * \param spectrum The spectral distribution for this road surface (if null, non draining default values will be used)
  * 
  * \returns true if all OK
  */
 bool NMPB08_DefineRoadSurfaceType(int idRoadSurface, 
	 double AVL, double BVL, double Vref_VL, 
	 double APL, double BPL, double Vref_PL,
	 double const* spectrum)
 {
	 if(idRoadSurface < RoadSurface_UserDefined) 
	 {
		 printf (".ERROR: definition of road surface type with idRoadSurface < RoadSurface_UserDefined \n") ;
		 return false;
	 }

	 return DefineRoadSurfaceType(idRoadSurface, 
		  AVL, BVL, Vref_VL, 
		  APL, BPL, Vref_PL,
		 spectrum);
 }
